GOOGLE_MODEL_TO_CONTEXT_LENGTH = {
    "gemini-2.5-pro-exp-03-25": 1048576,
    "gemini-2.0-flash-001": 1048576,
    "gemini-2.0-pro-exp-02-05": 2097152,
    "gemini-2.0-flash-lite-preview-02-05": 1048576,
    "gemini-2.0-flash-thinking-exp-01-21": 1048576,
    "gemini-1.5-flash": 1048576,
    "gemini-1.5-pro": 2097152,
    "gemini-1.0-pro": 32760,
    "gemini-1.0-pro-vision": 16384,
}

GOOGLE_MODEL_TO_OUTPUT_LENGTH = {"gemini-2.0-flash-001": 8192, "gemini-2.5-pro-exp-03-25": 65536}

GOOGLE_EMBEDING_MODEL_TO_DIM = {"text-embedding-005": 768, "text-multilingual-embedding-002": 768}
